from typing import Any, Optional

import numpy as np
from gymnasium.spaces import Space
from gymnasium.vector.utils import batch_space

from agilerl.typing import ActionType, PzStepReturn


class PettingZooVecEnv:
    """An abstract asynchronous, vectorized environment

    References:
        https://github.com/openai/baselines/tree/master/baselines/common/vec_env
        https://github.com/Farama-Foundation/Gymnasium/blob/main/gymnasium/vector/vector_env.py

    :param num_envs: Number of environments to vectorize
    :type num_envs: int
    :param observation_spaces: Dictionary of observation spaces
    :type observation_spaces: dict[str, gymnasium.spaces.Space]
    :param action_spaces: List of action spaces
    :type action_spaces: list[gymnasium.spaces.Space]
    :param possible_agents: List of possible agents
    :type possible_agents: list[str]
    """

    metadata: dict[str, Any] = {}
    render_mode: Optional[str] = None
    closed: bool = False
    num_envs: int
    agents: list[str]
    num_agents: int

    def __init__(
        self,
        num_envs: int,
        observation_spaces: dict[str, Space],
        action_spaces: list[Space],
        possible_agents: list[str],
    ):
        self.num_envs = num_envs
        self.agents = possible_agents
        self.num_agents = len(self.agents)
        self._single_observation_spaces = observation_spaces
        self._single_action_spaces = action_spaces
        self._observation_spaces = {
            agent: batch_space(space, self.num_envs)
            for agent, space in observation_spaces.items()
        }
        self._action_spaces = {
            agent: batch_space(space, self.num_envs)
            for agent, space in action_spaces.items()
        }
        self.action_space = self._get_action_space
        self.observation_space = self._get_observation_space
        self.single_action_space = self._get_single_action_space
        self.single_observation_space = self._get_single_observation_space

    def reset(
        self, seed: Optional[int] = None, options: Optional[dict[str, Any]] = None
    ) -> dict[str, Any]:
        """
        Reset all the environments and return two dictionaries of batched observations and infos.

        :param seed: Random seed, defaults to None
        :type seed: None | int, optional
        :param options: Options dictionary
        :type options: dict[str, Any]
        """
        pass

    def step_async(self, actions: list[dict[str, ActionType]]) -> None:
        """
        Tell all the environments to start taking a step
        with the given actions.
        Call step_wait() to get the results of the step.
        You should not call this if a step_async run is
        already pending.

        :param actions: List of dictionaries of length num_envs, each sub dictionary contains
        actions for each agent in a given environment
        :type actions: list[dict[str, int | float | np.ndarray]]
        """
        pass

    def step_wait(self) -> PzStepReturn:
        """
        Wait for the step taken with step_async().
        """
        pass

    def step(self, actions: dict[str, np.ndarray]) -> PzStepReturn:
        """Take an action for each parallel environment

        :param actions: Dictionary of vectorized actions for each agent.
        :type actions: dict[str, np.ndarray]

        :return: Tuple of observations, rewards, terminated, truncated, infos
        :rtype: tuple[dict[str, np.ndarray], dict[str, float], dict[str, bool], dict[str, bool], dict[str, Any]]
        """
        passed_actions_list = []
        for env_idx, _ in enumerate(list(actions.values())[0]):
            env_actions = {}
            for agent_id, action in actions.items():
                agent_action = action[env_idx]

                # Skip if the action is NaN
                if np.isnan(agent_action).all():
                    continue

                agent_action = (
                    int(agent_action) if np.isscalar(agent_action) else agent_action
                )
                env_actions[agent_id] = agent_action

            passed_actions_list.append(env_actions)
        assert (
            len(passed_actions_list) == self.num_envs
        ), "Number of actions passed to the step function must be equal to the number of vectorized environments"

        self.step_async(passed_actions_list)
        return self.step_wait()

    def render(self) -> Any:
        """Returns the rendered frames from the parallel environments."""
        raise NotImplementedError(
            f"{self.__str__()} render function is not implemented."
        )

    def close(self, **kwargs: Any) -> None:
        """
        Clean up the environments' resources.
        """
        if self.closed:
            return

        self.close_extras(**kwargs)
        self.closed = True

    def close_extras(self, **kwargs: Any) -> None:
        """Clean up the extra resources e.g. beyond what's in this base class."""
        pass

    @property
    def unwrapped(self) -> "PettingZooVecEnv":
        """Return the base environment."""
        return self

    def _get_single_action_space(self, agent):
        """Get an agents single action space

        :param agent: Name of agent
        :type agent: str
        """
        return self._single_action_spaces[agent]

    def _get_action_space(self, agent):
        """Get an agents action space

        :param agent: Name of agent
        :type agent: str
        """
        return self._action_spaces[agent]

    def _get_single_observation_space(self, agent):
        """Get an agents single observation space

        :param agent: Name of agent
        :type agent: str
        """
        return self._single_observation_spaces[agent]

    def _get_observation_space(self, agent):
        """Get an agents observation space

        :param agent: Name of agent
        :type agent: str
        """
        return self._observation_spaces[agent]



import gymnasium as gym

from agilerl.wrappers.learning import Skill


# Skill class wraps around an environment
def test_wrap_environment():
    env = gym.make("CartPole-v1")
    skill = Skill(env)

    assert isinstance(skill, Skill)
    assert isinstance(skill.env, gym.Wrapper)


# Skill class returns the skill_reward function output
def test_return_skill_reward():
    env = gym.make("CartPole-v1")

    class CustomSkill(Skill):
        def __init__(self, env):
            super().__init__(env)

        def skill_reward(self, observation, reward, terminated, truncated, info):
            return 0, 1, False, False, "Custom info"

    skill = CustomSkill(env)
    _, _ = env.reset()
    action = env.action_space.sample()
    new_observation, reward, terminated, truncated, info = skill.step(action)

    assert new_observation == 0
    assert reward == 1
    assert terminated is False
    assert truncated is False
    assert info == "Custom info"


# Skill class is initialized with a non-gym environment
def test_non_gym_environment():
    class CustomEnv(gym.Env):
        def __init__(self):
            self.observation_space = gym.spaces.Discrete(2)
            self.action_space = gym.spaces.Discrete(2)

        def reset(self):
            return self.observation_space.sample(), {}

        def step(self, action):
            return 0, 1, False, False, {}

    env = CustomEnv()
    skill = Skill(env)
    _, _ = env.reset()
    action = env.action_space.sample()
    new_observation, reward, terminated, truncated, info = skill.step(action)

    assert new_observation == 0
    assert reward == 1
    assert terminated is False
    assert truncated is False
    assert info == {}



import operator
from typing import Callable


class SegmentTree:
    """Create SegmentTree.

    Taken from OpenAI baselines github repository:
    https://github.com/openai/baselines/blob/master/baselines/common/segment_tree.py

    :param capacity: Capacity of segment tree
    :type capacity: int
    :param operation: Operation to apply
    :type operation: Callable
    :param init_value: Initial value
    :type init_value: float
    """

    def __init__(self, capacity: int, operation: Callable, init_value: float):
        assert (
            capacity > 0 and capacity & (capacity - 1) == 0
        ), "capacity must be positive and a power of 2."
        self.capacity = capacity
        self.tree = [init_value for _ in range(2 * capacity)]
        self.operation = operation

    def _operate_helper(
        self, start: int, end: int, node: int, node_start: int, node_end: int
    ) -> float:
        """Returns result of operation in segment.

        :param start: Start index of segment
        :type start: int
        :param end: End index of segment
        :type end: int
        :param node: Current node index
        :type node: int
        :param node_start: Start index of current node
        :type node_start: int
        :param node_end: End index of current node
        :type node_end: int

        :return: Result of operation in segment
        :rtype: float
        """
        if start == node_start and end == node_end:
            return self.tree[node]
        mid = (node_start + node_end) // 2
        if end <= mid:
            return self._operate_helper(start, end, 2 * node, node_start, mid)
        else:
            if mid + 1 <= start:
                return self._operate_helper(start, end, 2 * node + 1, mid + 1, node_end)
            else:
                return self.operation(
                    self._operate_helper(start, mid, 2 * node, node_start, mid),
                    self._operate_helper(mid + 1, end, 2 * node + 1, mid + 1, node_end),
                )

    def operate(self, start: int = 0, end: int = 0) -> float:
        """Returns result of applying `self.operation`.

        :param start: Start index of segment
        :type start: int
        :param end: End index of segment
        :type end: int

        :return: Result of applying `self.operation`
        :rtype: float
        """
        if end <= 0:
            end += self.capacity
        end -= 1

        return self._operate_helper(start, end, 1, 0, self.capacity - 1)

    def __setitem__(self, idx: int, val: float):
        """Set value in tree.

        :param idx: Index to set value at
        :type idx: int
        :param val: Value to set
        :type val: float
        """
        idx += self.capacity
        self.tree[idx] = val

        idx //= 2
        while idx >= 1:
            self.tree[idx] = self.operation(self.tree[2 * idx], self.tree[2 * idx + 1])
            idx //= 2

    def __getitem__(self, idx: int) -> float:
        """Get real value in leaf node of tree.

        :param idx: Index to get value at
        :type idx: int

        :return: Value at index
        :rtype: float
        """
        assert 0 <= idx < self.capacity

        return self.tree[self.capacity + idx]


class SumSegmentTree(SegmentTree):
    """Create SumSegmentTree.

    Taken from OpenAI baselines github repository:
    https://github.com/openai/baselines/blob/master/baselines/common/segment_tree.py

    :param capacity: Capacity of segment tree
    :type capacity: int
    """

    def __init__(self, capacity: int):
        super().__init__(capacity=capacity, operation=operator.add, init_value=0.0)

    def sum(self, start: int = 0, end: int = 0) -> float:
        """Returns sum of elements from start to end index.

        :param start: Start index of range, defaults to 0
        :type start: int, optional
        :param end: End index of range, defaults to 0 (meaning capacity)
        :type end: int, optional
        :return: Sum of elements in range [start, end)
        :rtype: float
        """
        return super().operate(start, end)

    def retrieve(self, upperbound: float) -> int:
        """Find the highest index `i` about `upperbound` in the tree

        :param upperbound: Upper bound for cumulative sum
        :type upperbound: float
        :return: Index where cumulative sum is <= upperbound
        :rtype: int
        """
        # TODO: Check assert case and fix bug
        assert 0 <= upperbound <= self.sum() + 1e-5, f"upperbound: {upperbound}"

        idx = 1
        while idx < self.capacity:  # while non-leaf
            left = 2 * idx
            right = left + 1
            if self.tree[left] > upperbound:
                idx = 2 * idx
            else:
                upperbound -= self.tree[left]
                idx = right
        return idx - self.capacity


class MinSegmentTree(SegmentTree):
    """Create SegmentTree.

    Taken from OpenAI baselines github repository:
    https://github.com/openai/baselines/blob/master/baselines/common/segment_tree.py

    :param capacity: Capacity of segment tree
    :type capacity: int
    """

    def __init__(self, capacity: int):
        super().__init__(capacity=capacity, operation=min, init_value=float("inf"))

    def min(self, start: int = 0, end: int = 0) -> float:
        """Returns minimum element from start to end index.

        :param start: Start index of range, defaults to 0
        :type start: int, optional
        :param end: End index of range, defaults to 0 (meaning capacity)
        :type end: int, optional
        :return: Minimum element in range [start, end)
        :rtype: float
        """
        return super().operate(start, end)



import copy
import warnings
from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import Any, Callable, Generator

import gymnasium as gym
import torch
import torch.nn as nn
from accelerate import Accelerator
from torch.utils.data import DataLoader

from agilerl import HAS_LLM_DEPENDENCIES
from agilerl.typing import PreferencePrompts, ReasoningPrompts

if HAS_LLM_DEPENDENCIES:
    import deepspeed
    from datasets import Dataset
    from transformers import AutoModelForCausalLM, AutoTokenizer
    from transformers.modeling_utils import PreTrainedModel
    from transformers.tokenization_utils_base import BatchEncoding

    AutoTokenizer = AutoTokenizer
else:
    AutoTokenizer = Any
    PreTrainedModel = Any
    BatchEncoding = Any
    Dataset = Any


def apply_chat_template(
    conversation_template: list[dict[str, str]],
    question: str,
    answer: str,
    tokenizer: AutoTokenizer,
) -> BatchEncoding:
    """
    Create and tokenize a chat template for a reaosning task.

    :param conversation_template: The conversation template to be tokenized.
    :type conversation_template: list[dict[str, str]]
    :param question: The question to be tokenized.
    :type question: str
    :param answer: The answer to be tokenized.
    :type answer: str
    :param tokenizer: The tokenizer to be used.
    :type tokenizer: AutoTokenizer
    :return: The tokenized prompt.
    :rtype: BatchEncoding
    """
    formatted_conversation = [
        {
            "role": msg["role"],
            "content": msg["content"].format(question=question, answer=answer),
        }
        for msg in conversation_template
    ]
    updated_prompt = tokenizer.apply_chat_template(
        formatted_conversation, tokenize=False, continue_final_message=True
    )
    tokenized_prompt = tokenizer(
        [updated_prompt],
        return_tensors="pt",
        padding=True,
        padding_side="left",
        return_attention_mask=True,
    )
    return tokenized_prompt


class HuggingFaceGym(gym.Env, ABC):
    """Abstract base class for HuggingFace Gymnasium environments.

    :param train_dataset: Train dataset to be loaded from HuggingFace datasets.
    :type train_dataset: Dataset
    :param test_dataset: Test dataset to be loaded from HuggingFace datasets.
    :type test_dataset: Dataset
    :param tokenizer: Tokenizer to be used for encoding and decoding the promåpts.
    :type tokenizer: AutoTokenizer
    :param custom_collate_fn: Custom collate function to be used for creating the batch, defaults to None
    :type custom_collate_fn: Callable, optional
    :param conversation_template: A structured conversation that acts as a base pattern for each data point.
    :type conversation_template: list[dict[str, str]]
    :param data_batch_size_per_gpu: DataLoader batch size, defaults to 8
    :type data_batch_size_per_gpu: int, optional
    :param max_context_length: Maximum context length, defaults to None
    :type max_context_length: int | None, optional
    :param min_completion_length: Minimum completion length, defaults to None
    :type min_completion_length: int, optional
    :param accelerator: Accelerator to be used for training, defaults to None
    :type accelerator: Accelerator, optional
    :param seed: Seed for the random number generator, defaults to 42
    :type seed: int, optional
    """

    def __init__(
        self,
        train_dataset: Dataset,
        test_dataset: Dataset,
        tokenizer: AutoTokenizer,
        conversation_template: list[dict[str, str]],
        data_batch_size_per_gpu: int = 8,
        max_context_length: int | None = None,
        min_completion_length: int = None,
        accelerator: Accelerator | None = None,
        seed: int = 42,
    ) -> None:
        self.name = train_dataset.info.dataset_name
        self.tokenizer = tokenizer
        self.data_batch_size_per_gpu = data_batch_size_per_gpu
        self.accelerator = accelerator
        self.min_completion_length = (
            0 if min_completion_length is None else min_completion_length
        )
        self.max_context_length = max_context_length
        self.seed = seed
        generator = torch.Generator().manual_seed(seed)
        self.conversation_template = conversation_template
        custom_collate_fn = self.create_collate_fn(tokenizer)
        dataloader_kwargs = {"collate_fn": custom_collate_fn}
        train_dataset = self._filter_dataset_by_max_context_length(
            train_dataset, "train dataset"
        )
        test_dataset = self._filter_dataset_by_max_context_length(
            test_dataset, "test dataset"
        )
        self.train_dataloader = DataLoader(
            train_dataset,
            batch_size=data_batch_size_per_gpu,
            shuffle=True,
            **dataloader_kwargs,
            generator=generator,
        )
        self.test_dataloader = DataLoader(
            test_dataset,
            batch_size=data_batch_size_per_gpu,
            shuffle=False,
            **dataloader_kwargs,
            generator=generator,
        )
        self.dataset_size = {
            "train": len(train_dataset),
            "test": len(test_dataset),
        }
        if self.accelerator is not None:
            self.train_dataloader = self.accelerator.prepare(self.train_dataloader)
            self.test_dataloader = self.accelerator.prepare(self.test_dataloader)
        self.train_dataloader_iter = iter(self.train_dataloader)
        self.test_dataloader_iter = iter(self.test_dataloader)
        self.dataloader = self.train_dataloader_iter
        self.reset_called = False
        self.evaluation_mode = False
        self.num_epochs = 0

    @abstractmethod
    def reset(
        self,
        reset_dataloaders: bool = False,
    ) -> tuple[list[ReasoningPrompts], dict[str, Any]]:
        """Reset the environment and get the next batch of tokenized prompts."""
        pass

    @abstractmethod
    def step(
        self,
        completions: torch.Tensor,
    ) -> tuple[list[ReasoningPrompts], torch.Tensor]:
        """Take a step in a HuggingFaceGym environment, calculate rewards from completions generated from previous prompt and provide new batch
        of prompts.
        """
        pass

    @contextmanager
    def eval_mode(self) -> Generator[None, None, None]:
        """Context manager to switch to evaluation mode."""
        self.dataloader = self.test_dataloader_iter
        self.evaluation_mode = True
        if hasattr(self, "last_tokenized_prompts"):
            last_tokenized_prompts = copy.deepcopy(self.last_tokenized_prompts)
        try:
            yield
        finally:
            self.dataloader = self.train_dataloader_iter
            self.evaluation_mode = False
            if hasattr(self, "last_tokenized_prompts"):
                self.last_tokenized_prompts = last_tokenized_prompts

    @abstractmethod
    def create_collate_fn(
        self,
        tokenizer: AutoTokenizer,
        *args,
        **kwargs,
    ) -> Callable[[list[dict[str, Any]]], dict[str, Any]]:
        """Create a collate function that applies the chat template to the batch of questions and answers."""
        pass

    def __len__(self):
        """Return the length of the dataset."""
        if self.evaluation_mode:
            return len(self.test_dataloader.dataset)
        return len(self.train_dataloader.dataset)

    def _reset_dataloaders(self, reset_train: bool = True, reset_test: bool = True):
        """Reset the dataloaders to the beginning of the dataset.

        :param reset_train: Whether to reset the train dataloader, defaults to True
        :type reset_train: bool, optional
        :param reset_test: Whether to reset the test dataloader, defaults to True
        :type reset_test: bool, optional
        """
        if reset_train:
            self.train_dataloader_iter = iter(self.train_dataloader)
        if reset_test:
            self.test_dataloader_iter = iter(self.test_dataloader)
        self.dataloader = (
            self.test_dataloader_iter
            if self.evaluation_mode
            else self.train_dataloader_iter
        )

    def _filter_dataset_by_max_context_length(
        self,
        dataset: Dataset,
        dataset_type: str | None = None,
    ) -> Dataset:
        """
        Filter the dataset by the max context length.

        :param dataset: Dataset to be filtered.
        :type dataset: Dataset
        :return: Filtered train and test datasets.
        :rtype: tuple[Dataset, Dataset]
        """
        dataset_type = "dataset" if dataset_type is None else dataset_type
        filter_keyword = "prompt" if "prompt" in dataset.features.keys() else "question"
        if self.max_context_length is None or not isinstance(
            dataset[0][filter_keyword], str
        ):
            return dataset
        filtered_dataset = dataset.filter(
            lambda x: len(self.tokenizer.encode(x[filter_keyword]))
            <= self.max_context_length - self.min_completion_length
        )
        if len(filtered_dataset) == 0:
            raise ValueError(
                f"No samples left in the {dataset_type} after filtering by the max context length constraint, use a larger max context length."
            )
        if (dataset_difference := len(dataset) - len(filtered_dataset)) > 0:
            warnings.warn(
                f"{dataset_difference} samples were filtered out of the {dataset_type} due to the max context length constraint."
            )
        return filtered_dataset


class ReasoningGym(HuggingFaceGym):
    """Class to convert HuggingFace datasets into Gymnasium style environment.

    :param train_dataset: Train dataset to be loaded from HuggingFace datasets.
    :type train_dataset: Dataset
    :param test_dataset: Test dataset to be loaded from HuggingFace datasets.
    :type test_dataset: Dataset
    :param tokenizer: Tokenizer to be used for encoding and decoding the prompts.
    :type tokenizer: AutoTokenizer
    :param reward_fn: Reward function for evaluating completions.
    :type reward_fn: Callable[..., float]
    :param conversation_template: A structured conversation that acts as a base pattern for each data point.
    :type conversation_template: list[dict[str, str]]
    :param data_batch_size_per_gpu: DataLoader batch size, defaults to 8
    :type data_batch_size_per_gpu: int, optional
    :param accelerator: Accelerator to be used for training, defaults to None
    :type accelerator: Accelerator, optional
    :param max_context_length: Maximum context length, defaults to None
    :type max_context_length: int | None, optional
    :param min_completion_length: Minimum completion length, defaults to 128
    :type min_completion_length: int, optional
    :param seed: Seed for the random number generator, defaults to 42
    :type seed: int, optional
    """

    def __init__(
        self,
        train_dataset: Dataset,
        test_dataset: Dataset,
        tokenizer: AutoTokenizer,
        reward_fn: Callable[[str, str, str], float],
        conversation_template: list[dict[str, str]],
        data_batch_size_per_gpu: int = 8,
        accelerator: Accelerator | None = None,
        return_raw_completions: bool = False,
        max_context_length: int | None = None,
        seed: int = 42,
    ) -> None:
        assert {"question", "answer"}.issubset(
            set(train_dataset.features.keys())
        ), "Train dataset must contain 'question' and 'answer' features."
        assert {"question", "answer"}.issubset(
            set(test_dataset.features.keys())
        ), "Train dataset must contain 'question' and 'answer' features."

        super().__init__(
            train_dataset=train_dataset,
            test_dataset=test_dataset,
            tokenizer=tokenizer,
            conversation_template=conversation_template,
            data_batch_size_per_gpu=data_batch_size_per_gpu,
            max_context_length=max_context_length,
            min_completion_length=0,
            accelerator=accelerator,
            seed=seed,
        )
        self.reward_fn = reward_fn
        self.return_raw_completions = return_raw_completions

    def step(
        self, completions: torch.Tensor
    ) -> tuple[list[ReasoningPrompts], torch.Tensor]:
        """Take a step in the ReasoningGym environment, calculate rewards from completions generated from previous prompt and provide new batch
        of prompts.

        :param completions: Completion IDs generated by the agent.
        :type completions: torch.Tensor
        :return: New tokenized prompts and an information dictionary.
        :rtype: tuple[list[BatchEncoding], torch.Tensor]
        """
        self.reset_called = False
        rewards = self._decode_and_evaluate(completions)
        new_tokenized_prompts = self._get_next_batch()
        self.last_tokenized_prompts = new_tokenized_prompts
        return new_tokenized_prompts, rewards

    def reset(
        self, reset_dataloaders: bool = False
    ) -> tuple[list[ReasoningPrompts], dict[str, Any]]:
        """Reset the environment and get the next batch of tokenized prompts.

        :param reset_dataloaders: Whether to reset the dataloaders, defaults to False
        :type reset_dataloaders: bool, optional
        :return: New tokenized prompts
        :rtype: tuple[list[BatchEncoding], dict[str, Any]]
        """
        if reset_dataloaders:
            self._reset_dataloaders()
            warnings.warn(
                "env.reset() called with reset_dataloaders=True, this will reset the dataloaders to the beginning of the dataset, proceed with caution."
            )
        if self.reset_called:
            warnings.warn(
                "env.reset() called more than once sequentially, it should typically follow with env.step()."
            )
        self.reset_called = True
        new_tokenized_prompts = self._get_next_batch()
        self.last_tokenized_prompts = new_tokenized_prompts
        return new_tokenized_prompts

    def _decode_and_evaluate(self, completions: list[torch.Tensor]) -> torch.Tensor:
        """Decode the completions and evaluate the rewards.

        :param completions: Completion IDs generated by the agent.
        :type completions: list[torch.Tensor]
        :return: Rewards for the completions.
        :rtype: torch.Tensor
        """
        # This is for a batch of completions (prompt_batch x group_size), List of tensors of length batch size, each tensor is a group of answers
        total_rewards = []
        for idx, (group_completion, answer, question) in enumerate(
            zip(completions, self.answers, self.questions)
        ):
            completion_to_decode = group_completion[
                :, self.last_tokenized_prompts[idx]["input_ids"].shape[1] :
            ]

            # Vectorize this in the future
            decoded_group_completion = self.tokenizer.batch_decode(
                completion_to_decode,
                skip_special_tokens=True,
            )
            rewards = [
                self.reward_fn(completion, answer, question)
                for completion in decoded_group_completion
            ]
            total_rewards.append(rewards)
        return torch.tensor(total_rewards)

    def _get_next_batch(self) -> list[ReasoningPrompts]:
        """Get the next batch of tokenized prompts."""
        try:
            batch = next(self.dataloader)
            self.questions = batch["question"]
            self.answers = batch["answer"]

            returned_prompts = [
                {
                    "input_ids": returned_prompt["input_ids"],
                    "attention_mask": returned_prompt["attention_mask"],
                    "text": (
                        self.tokenizer.batch_decode(
                            returned_prompt["input_ids"],
                            skip_special_tokens=False,  # Needs to be False here as we need to provide context about user roles to the model
                            clean_up_tokenization_spaces=False,
                        )[0]
                        if self.return_raw_completions
                        else None
                    ),
                }
                for returned_prompt in batch["tokenized_prompts"]
            ]
        except StopIteration:
            if not self.evaluation_mode:
                self.num_epochs += 1

            self._reset_dataloaders(
                reset_train=not self.evaluation_mode,
                reset_test=self.evaluation_mode,
            )
            return self._get_next_batch()
        return returned_prompts

    def create_collate_fn(
        self,
        tokenizer: AutoTokenizer,
    ) -> Callable[[list[dict[str, Any]]], dict[str, Any]]:
        """
        Create a collate function that applies the chat template to the batch of questions and answers.

        :param tokenizer: Tokenizer to be used for encoding and decoding the prompts.
        :type tokenizer: AutoTokenizer
        :return: Collate function that applies the chat template to the batch of questions and answers.
        :rtype: Callable[[list[dict[str, Any]]], dict[str, Any]]
        """

        def collate_fn(batch):
            questions = [item["question"] for item in batch]
            answers = [item["answer"] for item in batch]

            # Apply chat template to all samples
            tokenized_prompts = [
                apply_chat_template(self.conversation_template, q, a, tokenizer)
                for q, a in zip(questions, answers)
            ]

            return {
                "question": questions,
                "answer": answers,
                "tokenized_prompts": tokenized_prompts,  # Keep individual tokenized prompts
            }

        return collate_fn


class PreferenceGym(HuggingFaceGym):
    """Class to convert HuggingFace preference datasets into Gymnasium style environment.

    :param dataset_name: Dataset name to be loaded from HuggingFace datasets.
    :type dataset_name: str
    :param tokenizer: Tokenizer to be used for encoding and decoding the prompts.
    :type tokenizer: AutoTokenizer
    :param reward_fn: Reward function for evaluating completions.
    :type reward_fn: Callable[..., float]
    :param data_batch_size_per_gpu: DataLoader batch size, defaults to 8
    :type data_batch_size_per_gpu: int, optional
    :param custom_collate_fn: Custom collate function to be used for creating the batch, defaults to None
    :type custom_collate_fn: Callable, optional
    :param accelerator: Accelerator to be used for training, defaults to None
    :type accelerator: Accelerator, optional
    :param max_context_length: Maximum context length, defaults to None
    :type max_context_length: int | None, optional
    :param min_completion_length: Minimum completion length, defaults to 128
    :type min_completion_length: int, optional
    """

    def __init__(
        self,
        train_dataset: Dataset,
        test_dataset: Dataset,
        tokenizer: AutoTokenizer,
        data_batch_size_per_gpu: int = 8,
        accelerator: Accelerator | None = None,
        max_context_length: int | None = None,
        min_completion_length: int | None = None,
        seed: int = 42,
    ):
        super().__init__(
            train_dataset=train_dataset,
            test_dataset=test_dataset,
            tokenizer=tokenizer,
            conversation_template=None,
            data_batch_size_per_gpu=data_batch_size_per_gpu,
            max_context_length=max_context_length,
            min_completion_length=min_completion_length,
            accelerator=accelerator,
            seed=seed,
        )
        assert {"prompt", "chosen", "rejected"}.issubset(
            set(train_dataset.features.keys())
        ), "Train dataset must contain 'prompt', 'chosen', and 'rejected' features."
        assert {"prompt", "chosen", "rejected"}.issubset(
            set(test_dataset.features.keys())
        ), "Train dataset must contain 'prompt', 'chosen', and 'rejected' features."

    def reset(self, reset_dataloaders: bool = False) -> PreferencePrompts:
        """Reset the environment and get the next batch of tokenized prompts."""
        if reset_dataloaders:
            self._reset_dataloaders()
            warnings.warn(
                "env.reset() called with reset_dataloaders=True, this will reset the dataloaders to the beginning of the dataset, proceed with caution."
            )
        if self.reset_called:
            warnings.warn(
                "env.reset() called more than once sequentially, it should typically follow with env.step()."
            )
        self.reset_called = True
        return self._get_next_batch()

    def step(self) -> PreferencePrompts:
        """Take a step in the PreferenceGym environment, calculate rewards from completions generated from previous prompt and provide new batch
        of prompts.
        """
        self.reset_called = False
        return self._get_next_batch()

    def _get_next_batch(self) -> PreferencePrompts:
        try:
            batch = next(self.dataloader)
        except StopIteration:
            if not self.evaluation_mode:
                self.num_epochs += 1
            self._reset_dataloaders(
                reset_train=not self.evaluation_mode,
                reset_test=self.evaluation_mode,
            )
            return self._get_next_batch()
        return batch

    def create_collate_fn(
        self,
        tokenizer: AutoTokenizer,
        max_context_length: int | None = None,
    ) -> Callable[[list[dict[str, Any]]], dict[str, Any]]:
        """
        Create a collate function that applies the chat template to the batch of questions and answers.

        :param tokenizer: Tokenizer to be used for encoding and decoding the prompts.
        :type tokenizer: AutoTokenizer
        :param max_context_length: Maximum context length, defaults to None
        :type max_context_length: int | None, optional
        :return: Collate function that applies the chat template to the batch of questions and answers.
        :rtype: Callable[[list[dict[str, Any]]], dict[str, Any]]
        """

        def collate_fn(batch: list[dict[str, str]]) -> dict[str, str]:
            prompts = [item["prompt"] for item in batch]
            chosen = [item["chosen"] for item in batch]
            rejected = [item["rejected"] for item in batch]

            # Tokenize prompts separately to get their lengths
            prompt_encodings = tokenizer(
                prompts, truncation=True, padding=False, add_special_tokens=True
            )
            prompt_lengths = [len(ids) for ids in prompt_encodings["input_ids"]]

            # Tokenize without padding
            chosen_enc = tokenizer(
                prompts,
                chosen,
                max_length=self.max_context_length,
                truncation=True,
                padding=False,
            )
            rejected_enc = tokenizer(
                prompts,
                rejected,
                max_length=self.max_context_length,
                truncation=True,
                padding=False,
            )

            # Compute the joint max length across both
            max_len = max(
                max(len(ids) for ids in chosen_enc["input_ids"]),
                max(len(ids) for ids in rejected_enc["input_ids"]),
            )

            max_len = (
                min(max_len, self.max_context_length)
                if self.max_context_length is not None
                else max_len
            )

            # Now pad both encodings to the same target length
            chosen_enc = tokenizer.pad(
                chosen_enc,
                padding="max_length",
                max_length=max_len,
                return_tensors="pt",
            )
            rejected_enc = tokenizer.pad(
                rejected_enc,
                padding="max_length",
                max_length=max_len,
                return_tensors="pt",
            )

            return {
                "prompt": prompts,
                "prompt_lengths": prompt_lengths,
                "chosen": chosen,
                "rejected": rejected,
                "chosen_input_ids": chosen_enc["input_ids"],  # [batch_size, max_len]
                "chosen_attention_mask": chosen_enc[
                    "attention_mask"
                ].long(),  # [batch_size, max_len]
                "rejected_input_ids": rejected_enc[
                    "input_ids"
                ],  # [batch_size, max_len]
                "rejected_attention_mask": rejected_enc[
                    "attention_mask"
                ].long(),  # [batch_size, max_len]
            }

        return collate_fn


@contextmanager
def gather_if_zero3(
    zero_stage: int, params: list[torch.Tensor], modifier_rank: int | None = None
):
    """
    Conditional context manager for setting the zero stage for the model.

    :param zero_stage: The zero stage
    :type zero_stage: int
    :param params: The parameters to gather
    :type params: List[torch.Tensor]
    :param modifier_rank: The modifier rank
    :type modifier_rank: int | None
    """
    if zero_stage == 3:
        with deepspeed.zero.GatheredParameters(
            params=params,
            modifier_rank=modifier_rank,
        ):
            yield
    else:
        yield


def get_state_dict(model: nn.Module) -> dict[str, torch.Tensor]:
    """
    Get the state dict of the model for zero3.

    :param model: The model to get the state dict of.
    :type model: nn.Module
    :return: The state dict of the model.
    :rtype: dict[str, torch.Tensor]
    """

    with gather_if_zero3(3, list(model.parameters()), modifier_rank=0):
        return model.state_dict()


def create_model_from_name_or_path(
    model_name_or_path: str, model_config: dict[str, Any] | None = None
) -> PreTrainedModel:
    """
    Create a model from a name or path.

    :param model_name_or_path: The name or path of the model to create.
    :type model_name_or_path: str
    :param model_config: The configuration of the model to create.
    :type model_config: dict[str, Any ] | None
    :return: The created model.
    :rtype: PreTrainedModel
    """
    if model_config is None:
        model_config = {
            "torch_dtype": torch.bfloat16,
            "attn_implementation": "sdpa",
        }
    model = AutoModelForCausalLM.from_pretrained(
        pretrained_model_name_or_path=model_name_or_path, **model_config
    )
    return model



from typing import Any, Optional

import torch
import torch.nn as nn

from agilerl.modules.base import EvolvableModule, MutationType, mutation
from agilerl.typing import ArrayOrTensor, BatchDimension
from agilerl.utils.evolvable_networks import get_activation


class EvolvableLSTM(EvolvableModule):
    """The Evolvable Long Short-Term Memory (LSTM) class.

    :param input_size: Size of input features
    :type input_size: int
    :param hidden_state_size: Size of hidden state
    :type hidden_state_size: int
    :param num_outputs: Output dimension
    :type num_outputs: int
    :param num_layers: Number of LSTM layers stacked together
    :type num_layers: int
    :param output_activation: Output activation layer, defaults to None
    :type output_activation: str, optional
    :param min_hidden_state_size: Minimum hidden state size, defaults to 32
    :type min_hidden_state_size: int, optional
    :param max_hidden_state_size: Maximum hidden state size, defaults to 512
    :type max_hidden_state_size: int, optional
    :param min_layers: Minimum number of LSTM layers, defaults to 1
    :type min_layers: int, optional
    :param max_layers: Maximum number of LSTM layers, defaults to 3
    :type max_layers: int, optional
    :param dropout: Dropout probability between LSTM layers, defaults to 0.0
    :type dropout: float, optional
    :param device: Device for accelerated computing, 'cpu' or 'cuda', defaults to 'cpu'
    :type device: str, optional
    :param name: Name of the network, defaults to 'lstm'
    :type name: str, optional
    :param random_seed: Random seed to use for the network. Defaults to None.
    :type random_seed: Optional[int]
    """

    def __init__(
        self,
        input_size: int,
        hidden_state_size: int,
        num_outputs: int,
        num_layers: int = 1,
        output_activation: str = None,
        min_hidden_state_size: int = 32,
        max_hidden_state_size: int = 512,
        min_layers: int = 1,
        max_layers: int = 3,
        dropout: float = 0.0,
        device: str = "cpu",
        name: str = "lstm",
        random_seed: Optional[int] = None,
    ):
        super().__init__(device, random_seed)

        assert (
            input_size > 0
        ), "'input_size' cannot be less than or equal to zero, please enter a valid integer."
        assert (
            hidden_state_size > 0
        ), "'hidden_state_size' cannot be less than or equal to zero, please enter a valid integer."
        assert (
            num_outputs > 0
        ), "'num_outputs' cannot be less than or equal to zero, please enter a valid integer."
        assert (
            num_layers > 0
        ), "'num_layers' cannot be less than or equal to zero, please enter a valid integer."
        assert (
            min_hidden_state_size < max_hidden_state_size
        ), "'min_hidden_state_size' must be less than 'max_hidden_state_size'."
        assert min_layers < max_layers, "'min_layers' must be less than 'max_layers'."
        assert 0 <= dropout < 1, "'dropout' must be between 0 and 1."

        self.name = name
        self.input_size = input_size
        self.hidden_state_size = hidden_state_size
        self.num_outputs = num_outputs
        self.num_layers = num_layers
        self.output_activation = output_activation
        self.min_hidden_state_size = min_hidden_state_size
        self.max_hidden_state_size = max_hidden_state_size
        self.min_layers = min_layers
        self.max_layers = max_layers
        self.dropout = dropout

        # Create the network
        self.model = self.create_lstm()

    @property
    def hidden_state_architecture(self) -> dict[str, tuple[int, ...]]:
        """Returns the hidden state architecture."""
        # For LSTM, hidden state and cell state have shape (num_layers * num_directions, batch_size, hidden_size)
        # Assuming unidirectional LSTM (num_directions=1) !TODO: SHOULD WE HAVE A DIRECTIONAL LSTM IN THE FUTURE?
        return {
            "h": (self.num_layers, BatchDimension, self.hidden_state_size),
            "c": (self.num_layers, BatchDimension, self.hidden_state_size),
        }

    def create_lstm(self) -> nn.ModuleDict:
        """Creates and returns an LSTM network with the current configuration.

        :return: LSTM network
        :rtype: nn.ModuleDict
        """
        features_dim = int(self.hidden_state_size)

        # Define network components
        model_dict = nn.ModuleDict()
        model_dict[f"{self.name}_lstm"] = nn.LSTM(
            input_size=self.input_size,
            hidden_size=features_dim,
            num_layers=self.num_layers,
            batch_first=True,
            dropout=self.dropout if self.num_layers > 1 else 0,
            device=self.device,
        )

        # Add activation if specified
        model_dict[f"{self.name}_lstm_output"] = nn.Linear(
            features_dim, self.num_outputs, device=self.device
        )
        model_dict[f"{self.name}_output_activation"] = get_activation(
            self.output_activation
        )

        return model_dict

    @property
    def net_config(self) -> dict[str, Any]:
        """Returns model configuration in dictionary format."""
        net_config = self.init_dict.copy()
        for attr in ["input_size", "num_outputs", "device", "name"]:
            if attr in net_config:
                net_config.pop(attr)

        return net_config

    @property
    def activation(self) -> str:
        """Returns activation function."""
        return

    @activation.setter
    def activation(self, activation: str) -> None:
        """Set activation function."""
        pass

    @mutation(MutationType.ACTIVATION)
    def change_activation(self, activation: str, output: bool = False) -> None:
        """Set the output activation function for the network.

        :param activation: Activation function to use.
        :type activation: str
        :param output: Flag indicating whether to set the output activation function, defaults to False
        :type output: bool, optional
        """
        self.output_activation = activation
        self.recreate_network()

    def forward(
        self,
        x: ArrayOrTensor,
        hidden_state: Optional[dict[str, ArrayOrTensor]] = None,
    ) -> tuple[torch.Tensor, dict[str, torch.Tensor]]:
        """Forward pass of the network.

        :param x: Input tensor
        :type x: ArrayOrTensor
        :param hidden_state: Dict containing hidden and cell states, defaults to None
        :type hidden_state: Optional[dict[str, torch.Tensor]]
        :return: Output tensor and next hidden state
        :rtype: tuple[torch.Tensor, dict[str, torch.Tensor]]
        """
        if not isinstance(x, torch.Tensor):
            x = torch.tensor(x, dtype=torch.float32, device=self.device)

        # If input is 2D (batch_size, features), add sequence length dimension
        x_shape = x.shape
        if x.dim() == 2:
            x = x.unsqueeze(1)
        elif x.dim() != 3:
            raise ValueError(
                f"Expected 2D (batch_size, features) or 3D "
                f"(batch_size, seq_len, features) input, but got {x.dim()}D"
            )

        # Use provided hidden state if available
        if hidden_state is not None:
            h0 = hidden_state.get(f"{self.name}_h", None)
            c0 = hidden_state.get(f"{self.name}_c", None)

            # Reshape to (batch_seq_size, seq_len, features)
            sequence_input = False
            if h0.shape[1] != x_shape[0]:
                sequence_input = True
                x = x.view(h0.shape[1], -1, x_shape[1])

            lstm_output, (h_n, c_n) = self.model[f"{self.name}_lstm"](x, (h0, c0))

            # Reshape back to original shape
            lstm_output: torch.Tensor = lstm_output
            if sequence_input:
                out_shape = lstm_output.shape
                lstm_output = lstm_output.reshape(
                    out_shape[0] * out_shape[1], out_shape[2]
                )
            else:
                lstm_output = lstm_output.squeeze(1)
        else:
            raise ValueError("Hidden state is required for LSTM forward pass.")

        # Process output
        lstm_output = self.model[f"{self.name}_lstm_output"](lstm_output)
        lstm_output = self.model[f"{self.name}_output_activation"](lstm_output)

        # Return output and new hidden state
        next_hidden = {
            **hidden_state,
            f"{self.name}_h": h_n,
            f"{self.name}_c": c_n,
        }
        return lstm_output, next_hidden

    def get_output_dense(self) -> torch.nn.Module:
        """Returns output layer of neural network."""
        return self.model[f"{self.name}_linear_output"]

    @mutation(MutationType.LAYER)
    def add_layer(self) -> None:
        """Adds an LSTM layer to the network. Falls back on `add_node()` if
        max layers reached."""
        if self.num_layers < self.max_layers:  # HARD LIMIT
            self.num_layers += 1
        else:
            return self.add_node()

    @mutation(MutationType.LAYER)
    def remove_layer(self) -> None:
        """Removes an LSTM layer from the network. Falls back on `add_node()` if
        min layers reached."""
        if self.num_layers > self.min_layers:  # HARD LIMIT
            self.num_layers -= 1
        else:
            return self.add_node()

    @mutation(MutationType.NODE)
    def add_node(self, numb_new_nodes: Optional[int] = None) -> dict[str, int]:
        """Increases hidden size of the LSTM.

        :param numb_new_nodes: Number of nodes to add to hidden size, defaults to None
        :type numb_new_nodes: int, optional
        :return: Dictionary with number of new nodes
        :rtype: dict[str, int]
        """
        if numb_new_nodes is None:
            numb_new_nodes = self.rng.choice([16, 32, 64])

        if (
            self.hidden_state_size + numb_new_nodes <= self.max_hidden_state_size
        ):  # HARD LIMIT
            self.hidden_state_size += numb_new_nodes

        return {"numb_new_nodes": numb_new_nodes}

    @mutation(MutationType.NODE)
    def remove_node(self, numb_new_nodes: Optional[int] = None) -> dict[str, int]:
        """Decreases hidden size of the LSTM.

        :param numb_new_nodes: Number of nodes to remove from hidden size, defaults to None
        :type numb_new_nodes: int, optional
        :return: Dictionary with number of new nodes
        :rtype: dict[str, int]
        """
        if numb_new_nodes is None:
            numb_new_nodes = self.rng.choice([16, 32, 64])

        if (
            self.hidden_state_size - numb_new_nodes >= self.min_hidden_state_size
        ):  # HARD LIMIT
            self.hidden_state_size -= numb_new_nodes

        return {"numb_new_nodes": numb_new_nodes}

    def recreate_network(self) -> None:
        """Recreates the LSTM network with current parameters."""
        model = self.create_lstm()

        # Preserve parameters where possible
        self.model = EvolvableModule.preserve_parameters(
            old_net=self.model, new_net=model
        )



import copy

import pytest
import torch

from agilerl.modules.simba import EvolvableSimBa
from agilerl.networks import (
    ContinuousQNetwork,
    DeterministicActor,
    QNetwork,
    StochasticActor,
    ValueNetwork,
)
from tests.helper_functions import assert_state_dicts_equal


@pytest.fixture(scope="module")
def encoder_simba_config():
    return {
        "hidden_size": 64,
        "num_blocks": 3,
    }


def test_correct_initialization():
    model = EvolvableSimBa(
        num_inputs=10, num_outputs=2, hidden_size=64, num_blocks=3, device="cpu"
    )
    assert model.num_inputs == 10
    assert model.num_outputs == 2
    assert model.hidden_size == 64
    assert model.num_blocks == 3


def test_incorrect_initialization():
    with pytest.raises(TypeError):
        EvolvableSimBa(
            num_inputs=10,
            num_outputs=2,
            hidden_size=64,
            num_blocks="three",  # Incorrect type
            device="cpu",
        )


def test_forward_pass():
    model = EvolvableSimBa(
        num_inputs=10, num_outputs=2, hidden_size=64, num_blocks=3, device="cpu"
    )
    x = torch.randn(1, 10)
    output = model(x)
    assert output.shape == (1, 2)


def test_add_block():
    model = EvolvableSimBa(
        num_inputs=10, num_outputs=2, hidden_size=64, num_blocks=3, device="cpu"
    )
    initial_blocks = model.num_blocks
    model.add_block()
    assert model.num_blocks == initial_blocks + 1


def test_remove_block():
    model = EvolvableSimBa(
        num_inputs=10, num_outputs=2, hidden_size=64, num_blocks=3, device="cpu"
    )
    initial_blocks = model.num_blocks
    model.remove_block()
    assert model.num_blocks == initial_blocks - 1


@pytest.mark.parametrize("numb_new_nodes", [16, None])
def test_add_node(numb_new_nodes):
    model = EvolvableSimBa(
        num_inputs=10, num_outputs=2, hidden_size=64, num_blocks=3, device="cpu"
    )
    initial_hidden_size = model.hidden_size
    mut_dict = model.add_node(numb_new_nodes=numb_new_nodes)
    assert model.hidden_size == initial_hidden_size + mut_dict["numb_new_nodes"]


@pytest.mark.parametrize("numb_new_nodes", [16, None])
def test_remove_node(numb_new_nodes):
    model = EvolvableSimBa(
        num_inputs=10, num_outputs=2, hidden_size=124, num_blocks=3, device="cpu"
    )
    initial_hidden_size = model.hidden_size
    mut_dict = model.remove_node(numb_new_nodes=numb_new_nodes)
    assert model.hidden_size == initial_hidden_size - mut_dict["numb_new_nodes"]


def test_recreate_network():
    model = EvolvableSimBa(
        num_inputs=10, num_outputs=2, hidden_size=64, num_blocks=3, device="cpu"
    )
    initial_model = model.model
    model.recreate_network()
    assert model.model is not initial_model


def test_clone():
    model = EvolvableSimBa(
        num_inputs=10, num_outputs=2, hidden_size=64, num_blocks=3, device="cpu"
    )
    clone = model.clone()
    assert clone is not model
    assert clone.num_inputs == model.num_inputs
    assert clone.num_outputs == model.num_outputs
    assert clone.hidden_size == model.hidden_size
    assert clone.num_blocks == model.num_blocks
    assert_state_dicts_equal(clone.state_dict(), model.state_dict())


def test_deterministic_actor_simba(vector_space, encoder_simba_config):
    model = DeterministicActor(
        observation_space=vector_space,
        action_space=copy.deepcopy(vector_space),
        encoder_config=encoder_simba_config,
        simba=True,
        device="cpu",
    )
    assert isinstance(model.encoder, EvolvableSimBa)


def test_stochastic_actor_simba(vector_space, discrete_space, encoder_simba_config):
    model = StochasticActor(
        observation_space=vector_space,
        action_space=discrete_space,
        encoder_config=encoder_simba_config,
        simba=True,
        device="cpu",
    )
    assert isinstance(model.encoder, EvolvableSimBa)


def test_q_network_simba(vector_space, discrete_space, encoder_simba_config):
    model = QNetwork(
        observation_space=vector_space,
        action_space=discrete_space,
        encoder_config=encoder_simba_config,
        simba=True,
        device="cpu",
    )
    assert isinstance(model.encoder, EvolvableSimBa)


def test_continuous_q_network_simba(vector_space, encoder_simba_config):
    model = ContinuousQNetwork(
        observation_space=vector_space,
        action_space=copy.deepcopy(vector_space),
        encoder_config=encoder_simba_config,
        simba=True,
        device="cpu",
    )
    assert isinstance(model.encoder, EvolvableSimBa)


def test_value_network_simba(vector_space, encoder_simba_config):
    model = ValueNetwork(
        observation_space=vector_space,
        encoder_config=encoder_simba_config,
        simba=True,
        device="cpu",
    )
    assert isinstance(model.encoder, EvolvableSimBa)



from importlib.metadata import metadata
from importlib.util import find_spec

from packaging.requirements import Requirement


def get_extra_dependencies(package: str, extra: str) -> list[str]:
    requires = metadata(package).get_all("Requires-Dist") or []
    deps = []
    for req in requires:
        r = Requirement(req)
        if r.marker and r.marker.evaluate({"extra": extra}):
            deps.append(r.name)
    return deps


LLM_PACKAGES = get_extra_dependencies("agilerl", "llm")
HAS_LLM_DEPENDENCIES = all(find_spec(pkg) is not None for pkg in LLM_PACKAGES)



import warnings

import numpy as np
import pytest
import torch
from gymnasium import spaces
from tensordict import TensorDict

from agilerl.components.rollout_buffer import RolloutBuffer
from agilerl.typing import BPTTSequenceType


class TestRolloutBufferInitialization:
    """Test rollout buffer initialization with different configurations."""

    def test_init_box_observations(self):
        """Test initialization with Box observation space."""
        obs_space = spaces.Box(low=-1, high=1, shape=(4,), dtype=np.float32)
        action_space = spaces.Discrete(2)

        buffer = RolloutBuffer(
            capacity=10,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=2,
            device="cpu",
        )

        assert buffer.capacity == 10
        assert buffer.num_envs == 2
        assert buffer.pos == 0
        assert not buffer.full
        assert buffer.buffer is not None
        assert buffer.buffer["observations"].shape == (10, 2, 4)
        assert buffer.buffer["actions"].shape == (10, 2, 1)

    def test_init_discrete_observations(self):
        """Test initialization with Discrete observation space."""
        obs_space = spaces.Discrete(5)
        action_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)

        buffer = RolloutBuffer(
            capacity=5,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=3,
            device="cpu",
        )

        assert buffer.buffer["observations"].shape == (5, 3, 1)
        assert buffer.buffer["actions"].shape == (5, 3, 2)

    def test_init_dict_observations(self):
        """Test initialization with Dict observation space."""
        obs_space = spaces.Dict(
            {
                "vector": spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32),
                "discrete": spaces.Discrete(3),
            }
        )
        action_space = spaces.Discrete(2)

        buffer = RolloutBuffer(
            capacity=5,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=2,
            device="cpu",
        )

        assert isinstance(buffer.buffer["observations"], TensorDict)
        assert "vector" in buffer.buffer["observations"]
        assert "discrete" in buffer.buffer["observations"]
        assert buffer.buffer["observations"]["vector"].shape == (5, 2, 2)
        assert buffer.buffer["observations"]["discrete"].shape == (5, 2, 1)

    def test_init_recurrent_mode(self):
        """Test initialization with recurrent mode."""
        observation_space = spaces.Box(low=-1, high=1, shape=(4,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        buffer = RolloutBuffer(
            capacity=100,
            num_envs=1,
            observation_space=observation_space,
            action_space=action_space,
            device="cpu",
            gae_lambda=0.95,
            gamma=0.99,
        )

        assert buffer.capacity == 100
        assert buffer.observation_space == observation_space
        assert buffer.action_space == action_space
        assert buffer.gamma == 0.99
        assert buffer.gae_lambda == 0.95
        assert buffer.recurrent is False
        assert buffer.hidden_state_architecture is None
        assert buffer.device == "cpu"
        assert buffer.pos == 0
        assert buffer.full is False

        # Test with hidden states
        buffer = RolloutBuffer(
            capacity=100,
            num_envs=8,
            observation_space=observation_space,
            action_space=action_space,
            device="cpu",
            gae_lambda=0.95,
            gamma=0.99,
            recurrent=True,
            # (num_layers * directions, num_envs, hidden_size)
            hidden_state_architecture={
                "shared_encoder_h": (1, 8, 64),
                "shared_encoder_c": (1, 8, 64),
            },
        )

        assert buffer.recurrent is True
        assert buffer.hidden_state_architecture is not None
        assert buffer.buffer.get("hidden_states") is not None

    def test_init_error_recurrent_without_architecture(self):
        """Test error when recurrent=True but no hidden_state_architecture provided."""
        obs_space = spaces.Box(low=-1, high=1, shape=(4,), dtype=np.float32)
        action_space = spaces.Discrete(2)

        with pytest.raises(
            ValueError, match="hidden_state_architecture must be provided"
        ):
            RolloutBuffer(
                capacity=10,
                observation_space=obs_space,
                action_space=action_space,
                num_envs=2,
                recurrent=True,
                hidden_state_architecture=None,
            )


class TestRolloutBufferDataAddition:
    """Test adding data to the rollout buffer."""

    def test_add_box_observations(self):
        """Test adding data with Box observations."""
        observation_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        device = "cpu"

        buffer = RolloutBuffer(
            capacity=100,
            num_envs=1,
            observation_space=observation_space,
            action_space=action_space,
            device=device,
        )

        # Add a single sample
        obs = np.random.rand(*observation_space.shape).astype(observation_space.dtype)
        action = np.array([action_space.sample()])  # Ensure action is within space
        reward = 1.0
        done = False
        value = 0.5
        log_prob = -0.5
        next_obs = np.random.rand(*observation_space.shape).astype(
            observation_space.dtype
        )

        buffer.add(obs, action, reward, done, value, log_prob, next_obs)

        assert buffer.pos == 1
        assert not buffer.full
        # where X is the environment index, and Y is the position in the buffer (pos-1 for last added)
        # Data is stored at buffer.pos - 1
        current_pos_idx = buffer.pos - 1
        assert np.array_equal(
            buffer.buffer.get("observations")[current_pos_idx, 0].cpu().numpy(), obs
        )
        assert np.array_equal(
            buffer.buffer.get("actions")[current_pos_idx, 0, 0].cpu().numpy(), action[0]
        ), f"Expected action {action[0]} at position {current_pos_idx}, but got {buffer.buffer.get('actions')[current_pos_idx, 0].cpu().numpy()}"
        assert buffer.buffer.get("rewards")[current_pos_idx, 0].item() == reward
        assert buffer.buffer.get("dones")[current_pos_idx, 0].item() == float(done)
        assert buffer.buffer.get("values")[current_pos_idx, 0].item() == value
        assert buffer.buffer.get("log_probs")[current_pos_idx, 0].item() == log_prob
        assert np.array_equal(
            buffer.buffer.get("next_observations")[current_pos_idx, 0].cpu().numpy(),
            next_obs,
        )

        # Add samples until buffer is full
        for _ in range(buffer.capacity - 1):  # Already added one sample
            buffer.add(obs, action, reward, done, value, log_prob, next_obs)

        assert buffer.pos == buffer.capacity  # pos is next insertion point
        assert buffer.full is True  # Buffer is full when pos reaches capacity

        buffer.reset()

        assert buffer.pos == 0
        assert buffer.full is False

    def test_add_discrete_observations(self):
        """Test adding data with Discrete observations."""
        observation_space = spaces.Discrete(5)
        action_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        device = "cpu"
        num_envs = 8
        buffer = RolloutBuffer(
            capacity=100,
            num_envs=num_envs,
            observation_space=observation_space,
            action_space=action_space,
            device=device,
        )

        # Add a single sample
        obs = np.array([observation_space.sample() for _ in range(num_envs)])
        action = np.array([action_space.sample() for _ in range(num_envs)])
        reward = np.array([1.0 for _ in range(num_envs)])
        done = np.array([False for _ in range(num_envs)])
        value = np.array([0.5 for _ in range(num_envs)])
        log_prob = np.array([-0.5 for _ in range(num_envs)])
        next_obs = np.array([observation_space.sample() for _ in range(num_envs)])

        buffer.add(obs, action, reward, done, value, log_prob, next_obs)

        assert buffer.pos == 1
        assert not buffer.full
        assert buffer.buffer.get("observations").shape == (100, num_envs, 1)
        assert buffer.buffer.get("actions").shape == (100, num_envs, 2)
        assert buffer.buffer.get("rewards").shape == (100, num_envs)
        assert buffer.buffer.get("dones").shape == (100, num_envs)
        assert buffer.buffer.get("values").shape == (100, num_envs)
        assert buffer.buffer.get("log_probs").shape == (100, num_envs)
        assert buffer.buffer.get("next_observations").shape == (100, num_envs, 1)

    def test_add_multidiscrete_observations(self):
        """Test adding data with MultiDiscrete observations."""
        observation_space = spaces.MultiDiscrete([2, 3])
        action_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        device = "cpu"
        num_envs = 8
        buffer = RolloutBuffer(
            capacity=100,
            num_envs=num_envs,
            observation_space=observation_space,
            action_space=action_space,
            device=device,
        )

        # Add a single sample
        obs = np.array([observation_space.sample() for _ in range(num_envs)])
        action = np.array([action_space.sample() for _ in range(num_envs)])
        reward = np.array([1.0 for _ in range(num_envs)])
        done = np.array([False for _ in range(num_envs)])
        value = np.array([0.5 for _ in range(num_envs)])
        log_prob = np.array([-0.5 for _ in range(num_envs)])
        next_obs = np.array([observation_space.sample() for _ in range(num_envs)])

        buffer.add(obs, action, reward, done, value, log_prob, next_obs)

        assert buffer.pos == 1
        assert not buffer.full
        assert buffer.buffer.get("observations").shape == (100, num_envs, 2)
        assert buffer.buffer.get("actions").shape == (100, num_envs, 2)
        assert buffer.buffer.get("rewards").shape == (100, num_envs)
        assert buffer.buffer.get("dones").shape == (100, num_envs)
        assert buffer.buffer.get("values").shape == (100, num_envs)
        assert buffer.buffer.get("log_probs").shape == (100, num_envs)
        assert buffer.buffer.get("next_observations").shape == (100, num_envs, 2)

    def test_add_dict_observations(self):
        """Test adding data with Dict observations."""
        obs_space = spaces.Dict(
            {
                "vector": spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32),
                "discrete": spaces.Discrete(3),
            }
        )
        action_space = spaces.Discrete(2)

        buffer = RolloutBuffer(
            capacity=5,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=2,
            device="cpu",
        )

        obs = {
            "vector": np.array([[1.0, 2.0], [3.0, 4.0]]),
            "discrete": np.array([[0], [2]]),
        }
        action = np.array([0, 1])
        reward = np.array([1.0, -1.0])
        done = np.array([False, True])
        value = np.array([0.5, -0.5])
        log_prob = np.array([0.1, 0.2])

        buffer.add(
            obs=obs,
            action=action,
            reward=reward,
            done=done,
            value=value,
            log_prob=log_prob,
        )

        assert buffer.pos == 1
        assert isinstance(buffer.buffer["observations"], TensorDict)
        assert torch.allclose(
            buffer.buffer["observations"]["vector"][0],
            torch.tensor(obs["vector"], dtype=torch.float32),
        )
        assert torch.allclose(
            buffer.buffer["observations"]["discrete"][0],
            torch.tensor(obs["discrete"], dtype=torch.float32),
        )

    def test_add_with_next_obs(self):
        """Test adding data with next observations."""
        obs_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        buffer = RolloutBuffer(
            capacity=5,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=2,
            device="cpu",
        )

        obs = np.array([[1.0, 2.0], [3.0, 4.0]])
        next_obs = np.array([[1.1, 2.1], [3.1, 4.1]])
        action = np.array([0, 1])
        reward = np.array([1.0, -1.0])
        done = np.array([False, True])
        value = np.array([0.5, -0.5])
        log_prob = np.array([0.1, 0.2])

        buffer.add(
            obs=obs,
            action=action,
            reward=reward,
            done=done,
            value=value,
            log_prob=log_prob,
            next_obs=next_obs,
        )

        assert torch.allclose(
            buffer.buffer["next_observations"][0],
            torch.tensor(next_obs, dtype=torch.float32),
        )

    def test_add_recurrent_data(self):
        """Test adding data with hidden states."""
        obs_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        hidden_state_arch = {
            "h_actor": (1, 2, 4),  # (num_layers, num_envs, hidden_size)
        }

        buffer = RolloutBuffer(
            capacity=5,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=2,
            device="cpu",
            recurrent=True,
            hidden_state_architecture=hidden_state_arch,
        )

        obs = np.array([[1.0, 2.0], [3.0, 4.0]])
        action = np.array([0, 1])
        reward = np.array([1.0, -1.0])
        done = np.array([False, True])
        value = np.array([0.5, -0.5])
        log_prob = np.array([0.1, 0.2])
        hidden_state = {"h_actor": torch.randn(1, 2, 4)}

        buffer.add(
            obs=obs,
            action=action,
            reward=reward,
            done=done,
            value=value,
            log_prob=log_prob,
            hidden_state=hidden_state,
        )

        assert buffer.pos == 1
        assert buffer.recurrent
        assert "hidden_states" in buffer.buffer
        assert "h_actor" in buffer.buffer["hidden_states"]

    def test_add_capacity_overflow(self):
        """Test buffer behavior when capacity is exceeded."""
        obs_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        buffer = RolloutBuffer(
            capacity=2,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=1,
            device="cpu",
        )

        obs = np.array([[1.0, 2.0]])
        action = np.array([0])
        reward = np.array([1.0])
        done = np.array([False])
        value = np.array([0.5])
        log_prob = np.array([0.1])

        # Fill buffer to capacity
        buffer.add(obs, action, reward, done, value, log_prob)
        buffer.add(obs, action, reward, done, value, log_prob)

        # Should raise error when exceeding capacity
        with pytest.raises(ValueError, match="Buffer has reached capacity"):
            buffer.add(obs, action, reward, done, value, log_prob)

    def test_add_wrap_at_capacity(self):
        """Test buffer wrapping when wrap_at_capacity=True."""
        obs_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        buffer = RolloutBuffer(
            capacity=2,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=1,
            device="cpu",
            wrap_at_capacity=True,
        )

        obs = np.array([[1.0, 2.0]])
        action = np.array([0])
        reward = np.array([1.0])
        done = np.array([False])
        value = np.array([0.5])
        log_prob = np.array([0.1])

        # Fill buffer to capacity
        buffer.add(obs, action, reward, done, value, log_prob)
        buffer.add(obs, action, reward, done, value, log_prob)

        # Should wrap around
        buffer.add(obs, action, reward, done, value, log_prob)
        assert buffer.pos == 1
        assert buffer.full


class TestRolloutBufferReturnsAndAdvantages:
    """Test computation of returns and advantages."""

    def test_compute_returns_gae(self):
        """Test computing returns and advantages with GAE."""
        observation_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        device = "cpu"
        capacity = 5

        buffer = RolloutBuffer(
            capacity=capacity,
            num_envs=1,
            observation_space=observation_space,
            action_space=action_space,
            gamma=0.99,
            gae_lambda=0.95,
            use_gae=True,
            device=device,
        )

        # Add samples
        for i in range(capacity):
            obs = np.random.rand(*observation_space.shape).astype(
                observation_space.dtype
            )
            action = np.array([action_space.sample()])
            reward = 1.0
            done = i == (capacity - 1)  # Last step is done
            value = 0.5
            log_prob = -0.5
            next_obs = np.random.rand(*observation_space.shape).astype(
                observation_space.dtype
            )

            buffer.add(obs, action, reward, done, value, log_prob, next_obs)

        # Compute returns and advantages
        last_value = torch.tensor([[0.0]], device=device)  # Shape (num_envs, 1)
        last_done = torch.tensor([[0.0]], device=device)  # Shape (num_envs, 1)
        buffer.compute_returns_and_advantages(last_value, last_done)

        # Check that returns and advantages are computed
        # Slicing [:, 0] gets data for the first (and only) environment
        assert not np.array_equal(
            buffer.buffer.get("returns")[:, 0].cpu().numpy(), np.zeros((capacity, 1))
        )
        assert not np.array_equal(
            buffer.buffer.get("advantages")[:, 0].cpu().numpy(), np.zeros((capacity, 1))
        )

        # Check that returns are higher for earlier steps (due to discounting)
        assert (
            buffer.buffer.get("returns")[0, 0].item()
            > buffer.buffer.get("returns")[capacity - 1, 0].item()
        )

    def test_compute_returns_monte_carlo(self):
        """Test computing returns with Monte Carlo method."""
        obs_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        buffer = RolloutBuffer(
            capacity=5,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=2,
            device="cpu",
            use_gae=False,
            gamma=0.99,
        )

        # Add some data
        for i in range(3):
            obs = np.array([[i, i + 1], [i + 2, i + 3]])
            action = np.array([0, 1])
            reward = np.array([1.0, -1.0])
            done = np.array([False, False])
            value = np.array([0.5, -0.5])
            log_prob = np.array([0.1, 0.2])

            buffer.add(obs, action, reward, done, value, log_prob)

        last_value = np.array([0.8, 0.0])
        last_done = np.array([False, False])

        buffer.compute_returns_and_advantages(last_value, last_done)

        # Check that advantages and returns were computed
        advantages = buffer.buffer["advantages"][:3]
        returns = buffer.buffer["returns"][:3]

        assert advantages.shape == (3, 2)
        assert returns.shape == (3, 2)
        assert not torch.isnan(advantages).any()
        assert not torch.isnan(returns).any()

    def test_compute_returns_with_dones(self):
        """Test computing returns with done episodes."""
        obs_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        buffer = RolloutBuffer(
            capacity=5,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=2,
            device="cpu",
        )

        # Add data with some done episodes
        obs = np.array([[1.0, 2.0], [3.0, 4.0]])
        action = np.array([0, 1])
        reward = np.array([1.0, -1.0])
        done = np.array([False, True])  # Second env episode ends
        value = np.array([0.5, -0.5])
        log_prob = np.array([0.1, 0.2])

        buffer.add(obs, action, reward, done, value, log_prob)

        last_value = np.array([0.8, 0.0])
        last_done = np.array([False, True])

        buffer.compute_returns_and_advantages(last_value, last_done)

        advantages = buffer.buffer["advantages"][:1]
        returns = buffer.buffer["returns"][:1]

        assert advantages.shape == (1, 2)
        assert returns.shape == (1, 2)
        assert not torch.isnan(advantages).any()
        assert not torch.isnan(returns).any()


class TestRolloutBufferDataRetrieval:
    """Test retrieving data from the buffer."""

    def test_get_numpy_data(self):
        """Test getting data as numpy arrays."""
        obs_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        buffer = RolloutBuffer(
            capacity=5,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=2,
            device="cpu",
        )

        # Add test data
        obs = np.array([[1.0, 2.0], [3.0, 4.0]])
        action = np.array([0, 1])
        reward = np.array([1.0, -1.0])
        done = np.array([False, True])
        value = np.array([0.5, -0.5])
        log_prob = np.array([0.1, 0.2])

        buffer.add(obs, action, reward, done, value, log_prob)

        # Compute returns
        last_value = np.array([0.8, 0.0])
        last_done = np.array([False, True])
        buffer.compute_returns_and_advantages(last_value, last_done)

        # Get data
        data = buffer.get()

        assert isinstance(data, dict)
        assert "observations" in data
        assert "actions" in data
        assert "rewards" in data
        assert "advantages" in data
        assert "returns" in data

        # Check shapes and types
        assert isinstance(data["observations"], np.ndarray)
        assert data["observations"].shape == (2, 2)  # 2 samples, 2 features
        assert data["actions"].shape == (2, 1)

    def test_get_tensor_batch(self):
        """Test getting tensor batch."""
        observation_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        device = "cpu"
        num_samples = 10

        buffer = RolloutBuffer(
            capacity=100,
            num_envs=1,
            observation_space=observation_space,
            action_space=action_space,
            device=device,
        )

        # Add samples
        for i in range(num_samples):
            obs = np.random.rand(*observation_space.shape).astype(
                observation_space.dtype
            )
            action = np.array([action_space.sample()])
            reward = 1.0
            done = i == (num_samples - 1)  # Last step is done
            value = 0.5
            log_prob = -0.5
            next_obs = np.random.rand(*observation_space.shape).astype(
                observation_space.dtype
            )

            buffer.add(obs, action, reward, done, value, log_prob, next_obs)

        # Compute returns and advantages
        last_value = torch.tensor([[0.0]], device=device)
        last_done = torch.tensor([[0.0]], device=device)
        buffer.compute_returns_and_advantages(last_value, last_done)

        # Get all data (up to current pos)
        batch = buffer.get()  # Gets all data up to buffer.pos

        assert len(batch["observations"]) == num_samples
        assert len(batch["actions"]) == num_samples
        # Rewards, dones, values, log_probs are (num_samples, 1) after get() flattens num_envs
        assert len(batch["rewards"]) == num_samples
        assert len(batch["dones"]) == num_samples
        assert len(batch["values"]) == num_samples
        assert len(batch["log_probs"]) == num_samples
        assert len(batch["advantages"]) == num_samples
        assert len(batch["returns"]) == num_samples

        # Get batch of specific size
        batch_size = 5
        batch = buffer.get(batch_size=batch_size)

        assert len(batch["observations"]) == batch_size
        assert len(batch["actions"]) == batch_size

        # Get tensor batch
        tensor_batch = buffer.get_tensor_batch(batch_size=batch_size)

        assert isinstance(tensor_batch["observations"], torch.Tensor)
        assert isinstance(tensor_batch["actions"], torch.Tensor)
        assert isinstance(tensor_batch["advantages"], torch.Tensor)
        assert tensor_batch["observations"].shape[0] == batch_size

    def test_get_dict_observations(self):
        """Test getting data with dict observations."""
        obs_space = spaces.Dict(
            {
                "vector": spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32),
                "discrete": spaces.Discrete(3),
            }
        )
        action_space = spaces.Discrete(2)

        buffer = RolloutBuffer(
            capacity=5,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=2,
            device="cpu",
        )

        obs = {
            "vector": np.array([[1.0, 2.0], [3.0, 4.0]]),
            "discrete": np.array([[0], [1]]),
        }
        action = np.array([0, 1])
        reward = np.array([1.0, -1.0])
        done = np.array([False, True])
        value = np.array([0.5, -0.5])
        log_prob = np.array([0.1, 0.2])

        buffer.add(obs, action, reward, done, value, log_prob)

        # Compute returns
        last_value = np.array([0.8, 0.0])
        last_done = np.array([False, True])
        buffer.compute_returns_and_advantages(last_value, last_done)

        # Get data
        data = buffer.get()

        assert isinstance(data["observations"], dict)
        assert "vector" in data["observations"]
        assert "discrete" in data["observations"]
        assert data["observations"]["vector"].shape == (2, 2)
        assert data["observations"]["discrete"].shape == (2, 1)

    def test_get_with_batch_size(self):
        """Test getting data with specific batch size."""
        obs_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        buffer = RolloutBuffer(
            capacity=10,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=2,
            device="cpu",
        )

        # Add multiple timesteps
        for i in range(3):
            obs = np.array([[i, i + 1], [i + 2, i + 3]])
            action = np.array([0, 1])
            reward = np.array([1.0, -1.0])
            done = np.array([False, False])
            value = np.array([0.5, -0.5])
            log_prob = np.array([0.1, 0.2])

            buffer.add(obs, action, reward, done, value, log_prob)

        # Compute returns
        last_value = np.array([0.8, 0.0])
        last_done = np.array([False, False])
        buffer.compute_returns_and_advantages(last_value, last_done)

        # Get data with batch size
        data = buffer.get(batch_size=4)

        assert data["observations"].shape[0] == 4
        assert data["actions"].shape[0] == 4


class TestRolloutBufferSequences:
    """Test sequence methods for BPTT."""

    def test_prepare_sequence_tensors_basic(self):
        """Test basic sequence tensor preparation."""
        obs_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        hidden_state_arch = {
            "h_actor": (1, 2, 4),
        }

        buffer = RolloutBuffer(
            capacity=10,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=2,
            device="cpu",
            recurrent=True,
            hidden_state_architecture=hidden_state_arch,
            max_seq_len=3,
        )

        # Add multiple timesteps with episodes ending
        for i in range(5):
            obs = np.array([[i, i + 1], [i + 2, i + 3]])
            action = np.array([i % 2, (i + 1) % 2])
            reward = np.array([1.0, -1.0])
            done = np.array([i == 4, i == 2])  # Episodes end at different times
            value = np.array([0.5, -0.5])
            log_prob = np.array([0.1, 0.2])
            hidden_state = {"h_actor": torch.randn(1, 2, 4)}

            buffer.add(
                obs, action, reward, done, value, log_prob, hidden_state=hidden_state
            )

        # Compute returns
        last_value = np.array([0.8, 0.0])
        last_done = np.array([True, True])
        buffer.compute_returns_and_advantages(last_value, last_done)

        # Prepare sequence tensors
        buffer.prepare_sequence_tensors()

        # Check that sequence data was prepared
        assert buffer.padded_data is not None
        assert buffer.unpadded_data is not None
        assert buffer.unpadded_slices is not None
        assert buffer.num_sequences is not None
        assert buffer.max_sequence_length is not None

        # Check that padded data contains expected keys
        assert "observations" in buffer.padded_data
        assert "actions" in buffer.padded_data
        assert "pad_mask" in buffer.padded_data

        # Check that unpadded data contains expected keys
        assert "log_probs" in buffer.unpadded_data
        assert "advantages" in buffer.unpadded_data
        assert "values" in buffer.unpadded_data
        assert "returns" in buffer.unpadded_data

    @pytest.mark.parametrize("recurrent", [True, False])
    def test_prepare_sequence_tensors_with_hidden_states(self, recurrent):
        """Test sequence tensor preparation with hidden states."""
        obs_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        hidden_state_arch = {
            "h_actor": (1, 2, 4),
            "h_critic": (2, 2, 8),
        }

        buffer = RolloutBuffer(
            capacity=10,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=2,
            device="cpu",
            recurrent=recurrent,
            hidden_state_architecture=hidden_state_arch,
            max_seq_len=4,
        )

        # Add multiple timesteps
        for i in range(6):
            obs = np.array([[i, i + 1], [i + 2, i + 3]])
            action = np.array([i % 2, (i + 1) % 2])
            reward = np.array([1.0, -1.0])
            done = np.array([i == 5, i == 3])  # Episodes end at different times
            value = np.array([0.5, -0.5])
            log_prob = np.array([0.1, 0.2])
            hidden_state = {
                "h_actor": torch.randn(1, 2, 4),
                "h_critic": torch.randn(2, 2, 8),
            }

            buffer.add(
                obs, action, reward, done, value, log_prob, hidden_state=hidden_state
            )

        # Compute returns
        last_value = np.array([0.8, 0.0])
        last_done = np.array([True, True])
        buffer.compute_returns_and_advantages(last_value, last_done)

        # Prepare sequence tensors
        if recurrent:
            buffer.prepare_sequence_tensors()
        else:
            with pytest.raises(ValueError):
                buffer.prepare_sequence_tensors()
            return

        # Check that initial hidden states are prepared
        try:
            initial_hidden_states = buffer.padded_data.get_non_tensor(
                "initial_hidden_states"
            )
            assert "h_actor" in initial_hidden_states
            assert "h_critic" in initial_hidden_states

            # Check shapes of initial hidden states
            assert initial_hidden_states["h_actor"].shape[1:] == (
                1,
                4,
            )  # (batch, layers, hidden_size)
            assert initial_hidden_states["h_critic"].shape[1:] == (2, 8)
        except KeyError:
            # If initial_hidden_states is not found, the test should fail
            assert False, "initial_hidden_states should be present for recurrent buffer"

    def test_get_minibatch_sequences(self):
        """Test getting minibatch sequences."""
        obs_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        hidden_state_arch = {
            "h_actor": (1, 2, 4),
        }

        buffer = RolloutBuffer(
            capacity=15,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=2,
            device="cpu",
            recurrent=True,
            hidden_state_architecture=hidden_state_arch,
            max_seq_len=3,
        )

        # Add multiple episodes
        for i in range(10):
            obs = np.array([[i, i + 1], [i + 2, i + 3]])
            action = np.array([i % 2, (i + 1) % 2])
            reward = np.array([1.0, -1.0])
            # Create episodes of different lengths
            done = np.array([i % 4 == 3, i % 5 == 4])
            value = np.array([0.5, -0.5])
            log_prob = np.array([0.1, 0.2])
            hidden_state = {"h_actor": torch.randn(1, 2, 4)}

            buffer.add(
                obs, action, reward, done, value, log_prob, hidden_state=hidden_state
            )

        # Compute returns
        last_value = np.array([0.8, 0.0])
        last_done = np.array([True, True])
        buffer.compute_returns_and_advantages(last_value, last_done)

        # Prepare sequence tensors
        buffer.prepare_sequence_tensors()

        # Get minibatch sequences
        batch_size = 2
        minibatch_generator = buffer.get_minibatch_sequences(batch_size)

        # Test that we can iterate through minibatches
        minibatch_count = 0
        for padded_batch, unpadded_batch in minibatch_generator:
            minibatch_count += 1

            # Check that batches are TensorDicts
            assert isinstance(padded_batch, TensorDict)
            assert isinstance(unpadded_batch, TensorDict)

            # Check that padded batch contains expected keys
            assert "observations" in padded_batch
            assert "actions" in padded_batch
            assert "pad_mask" in padded_batch

            # Check that unpadded batch contains expected keys
            assert "log_probs" in unpadded_batch
            assert "advantages" in unpadded_batch
            assert "values" in unpadded_batch
            assert "returns" in unpadded_batch

            # Check that initial hidden states are present
            try:
                initial_hidden_states = padded_batch.get_non_tensor(
                    "initial_hidden_states"
                )
                assert "h_actor" in initial_hidden_states
            except KeyError:
                # It's okay if initial_hidden_states is not present for this test
                pass

        assert minibatch_count > 0

    def test_get_minibatch_sequences_without_preparation(self):
        """Test that get_minibatch_sequences raises error without preparation."""
        obs_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)

        buffer = RolloutBuffer(
            capacity=10,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=2,
            device="cpu",
            max_seq_len=3,
        )

        # Try to get minibatch sequences without preparing
        with pytest.raises(
            ValueError,
            match="Attempting to fetch minibatches before preparing sequences",
        ):
            list(buffer.get_minibatch_sequences(2))

    def test_pad_sequences_static_method(self):
        """Test the static _pad_sequences method."""
        # Test with regular tensors
        sequences = [
            torch.tensor([[1, 2], [3, 4]]),  # Length 2
            torch.tensor([[5, 6], [7, 8], [9, 10]]),  # Length 3
            torch.tensor([[11, 12]]),  # Length 1
        ]

        # Test padding to max length
        padded = RolloutBuffer._pad_sequences(sequences)
        assert padded.shape == (3, 3, 2)  # (batch_size, max_seq_len, feature_dim)

        # Test padding to specific target length
        padded_target = RolloutBuffer._pad_sequences(sequences, target_length=5)
        assert padded_target.shape == (
            3,
            5,
            2,
        )  # (batch_size, target_length, feature_dim)

        # Test with TensorDict sequences
        td_sequences = [
            TensorDict(
                {"obs": torch.tensor([[1, 2]]), "act": torch.tensor([0])},
                batch_size=[1],
            ),
            TensorDict(
                {"obs": torch.tensor([[3, 4], [5, 6]]), "act": torch.tensor([1, 0])},
                batch_size=[2],
            ),
        ]

        padded_td = RolloutBuffer._pad_sequences(td_sequences)
        assert isinstance(padded_td, TensorDict)
        assert padded_td["obs"].shape == (
            2,
            2,
            2,
        )  # (batch_size, max_seq_len, feature_dim)
        assert padded_td["act"].shape == (2, 2)  # (batch_size, max_seq_len)

    def test_different_bptt_sequence_types(self):
        """Test different BPTT sequence types."""
        from agilerl.typing import BPTTSequenceType

        obs_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        hidden_state_arch = {
            "h_actor": (1, 2, 4),
        }

        max_seq_len = 4
        episode_length = 12
        for seq_type in [
            BPTTSequenceType.CHUNKED,
            BPTTSequenceType.MAXIMUM,
            BPTTSequenceType.FIFTY_PERCENT_OVERLAP,
        ]:
            buffer = RolloutBuffer(
                capacity=20,
                observation_space=obs_space,
                action_space=action_space,
                num_envs=1,
                device="cpu",
                max_seq_len=max_seq_len,
                bptt_sequence_type=seq_type,
                recurrent=True,
                hidden_state_architecture=hidden_state_arch,
            )

            # Add a long episode
            for i in range(episode_length):
                obs = np.array([[i, i + 1]])
                action = np.array([i % 2])
                reward = np.array([1.0])
                done = np.array([i == 11])  # Episode ends at the last step
                value = np.array([0.5])
                log_prob = np.array([0.1])

                buffer.add(obs, action, reward, done, value, log_prob)

            # Compute returns
            last_value = np.array([0.0])
            last_done = np.array([True])
            buffer.compute_returns_and_advantages(last_value, last_done)

            # Prepare sequence tensors
            buffer.prepare_sequence_tensors()

            # Check that the number of sequences is correct
            if seq_type == BPTTSequenceType.CHUNKED:
                assert buffer.num_sequences == episode_length // max_seq_len
            elif seq_type == BPTTSequenceType.MAXIMUM:
                assert buffer.num_sequences == episode_length - max_seq_len + 1
            elif seq_type == BPTTSequenceType.FIFTY_PERCENT_OVERLAP:
                assert buffer.num_sequences == 5

            # Check that sequences were created
            assert buffer.num_sequences > 0
            assert buffer.max_sequence_length <= 4

    def test_empty_buffer_sequence_preparation(self):
        """Test sequence preparation with empty buffer."""
        obs_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        hidden_state_arch = {
            "h_actor": (1, 2, 4),
        }

        buffer = RolloutBuffer(
            capacity=10,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=2,
            device="cpu",
            max_seq_len=3,
            recurrent=True,
            hidden_state_architecture=hidden_state_arch,
        )

        # Try to prepare sequences with empty buffer
        with pytest.raises(ValueError):
            buffer.prepare_sequence_tensors()

    def test_sequence_preparation_with_dict_observations(self):
        """Test sequence preparation with dictionary observations."""
        obs_space = spaces.Dict(
            {
                "vector": spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32),
                "discrete": spaces.Discrete(3),
            }
        )
        action_space = spaces.Discrete(2)
        hidden_state_arch = {
            "h_actor": (1, 2, 4),
        }

        buffer = RolloutBuffer(
            capacity=10,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=2,
            device="cpu",
            max_seq_len=3,
            recurrent=True,
            hidden_state_architecture=hidden_state_arch,
        )

        # Add multiple timesteps
        for i in range(5):
            obs = {
                "vector": np.array([[i, i + 1], [i + 2, i + 3]]),
                "discrete": np.array([[i % 3], [(i + 1) % 3]]),
            }
            action = np.array([i % 2, (i + 1) % 2])
            reward = np.array([1.0, -1.0])
            done = np.array([i == 4, i == 2])
            value = np.array([0.5, -0.5])
            log_prob = np.array([0.1, 0.2])

            buffer.add(obs, action, reward, done, value, log_prob)

        # Compute returns
        last_value = np.array([0.8, 0.0])
        last_done = np.array([True, True])
        buffer.compute_returns_and_advantages(last_value, last_done)

        # Prepare sequence tensors
        buffer.prepare_sequence_tensors()

        # Check that observations are properly structured
        assert "observations" in buffer.padded_data
        obs_data = buffer.padded_data["observations"]
        assert isinstance(obs_data, TensorDict)
        assert "vector" in obs_data
        assert "discrete" in obs_data

    def test_sequence_preparation_different_max_seq_len(self):
        """Test sequence preparation with different max_seq_len values."""
        obs_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        hidden_state_arch = {
            "h_actor": (1, 2, 4),
        }
        # Test with max_seq_len=None (should use full episodes)
        buffer_no_limit = RolloutBuffer(
            capacity=10,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=1,
            device="cpu",
            max_seq_len=None,
            recurrent=True,
            hidden_state_architecture=hidden_state_arch,
        )

        # Add a long episode
        for i in range(8):
            obs = np.array([[i, i + 1]])
            action = np.array([i % 2])
            reward = np.array([1.0])
            done = np.array([i == 7])  # Episode ends at the last step
            value = np.array([0.5])
            log_prob = np.array([0.1])

            buffer_no_limit.add(obs, action, reward, done, value, log_prob)

        # Compute returns
        last_value = np.array([0.0])
        last_done = np.array([True])
        buffer_no_limit.compute_returns_and_advantages(last_value, last_done)

        # Prepare sequence tensors
        buffer_no_limit.prepare_sequence_tensors()

        # Should have one sequence of length 8
        assert buffer_no_limit.num_sequences == 1
        assert buffer_no_limit.max_sequence_length == 8

    def test_minibatch_sequences_batch_size_larger_than_sequences(self):
        """Test minibatch sequences when batch size is larger than available sequences."""
        obs_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        hidden_state_arch = {
            "h_actor": (1, 2, 4),
        }
        buffer = RolloutBuffer(
            capacity=10,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=1,
            device="cpu",
            max_seq_len=3,
            recurrent=True,
            hidden_state_architecture=hidden_state_arch,
        )

        # Add a short episode
        for i in range(3):
            obs = np.array([[i, i + 1]])
            action = np.array([i % 2])
            reward = np.array([1.0])
            done = np.array([i == 2])
            value = np.array([0.5])
            log_prob = np.array([0.1])

            buffer.add(obs, action, reward, done, value, log_prob)

        # Compute returns
        last_value = np.array([0.0])
        last_done = np.array([True])
        buffer.compute_returns_and_advantages(last_value, last_done)

        # Prepare sequence tensors
        buffer.prepare_sequence_tensors()

        # Request batch size larger than available sequences
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            minibatch_generator = buffer.get_minibatch_sequences(batch_size=10)
            list(minibatch_generator)  # Consume the generator

            # Should have issued a warning
            assert len(w) > 0
            assert "larger than the number of sequences" in str(w[0].message)

    def test_bptt_sequence_type_parameter(self):
        """Test that bptt_sequence_type parameter is properly stored."""
        from agilerl.typing import BPTTSequenceType

        obs_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        hidden_state_arch = {
            "h_actor": (1, 2, 4),
        }
        buffer = RolloutBuffer(
            capacity=10,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=1,
            device="cpu",
            max_seq_len=3,
            bptt_sequence_type=BPTTSequenceType.MAXIMUM,
            recurrent=True,
            hidden_state_architecture=hidden_state_arch,
        )

        assert buffer.bptt_sequence_type == BPTTSequenceType.MAXIMUM

    def test_sequence_preparation_device_parameter(self):
        """Test sequence preparation with specific device parameter."""
        obs_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        hidden_state_arch = {
            "h_actor": (1, 2, 4),
        }
        buffer = RolloutBuffer(
            capacity=10,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=1,
            device="cpu",
            max_seq_len=3,
            recurrent=True,
            hidden_state_architecture=hidden_state_arch,
        )

        # Add some data
        for i in range(3):
            obs = np.array([[i, i + 1]])
            action = np.array([i % 2])
            reward = np.array([1.0])
            done = np.array([i == 2])
            value = np.array([0.5])
            log_prob = np.array([0.1])

            buffer.add(obs, action, reward, done, value, log_prob)

        # Compute returns
        last_value = np.array([0.0])
        last_done = np.array([True])
        buffer.compute_returns_and_advantages(last_value, last_done)

        # Prepare sequence tensors with specific device
        buffer.prepare_sequence_tensors(device="cpu")

        # Check that data is on the correct device
        assert buffer.padded_data.device.type == "cpu"
        assert buffer.unpadded_data.device.type == "cpu"


class TestRolloutBufferUtilities:
    """Test utility methods and edge cases."""

    @pytest.mark.parametrize(
        "bptt_sequence_type",
        [
            BPTTSequenceType.CHUNKED,
            BPTTSequenceType.MAXIMUM,
            BPTTSequenceType.FIFTY_PERCENT_OVERLAP,
        ],
    )
    @pytest.mark.parametrize("max_seq_len", [None, 3, 5])
    def test_get_complete_sequences_method(self, bptt_sequence_type, max_seq_len):
        """Test the _get_complete_sequences method directly."""
        obs_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        hidden_state_arch = {
            "h_actor": (1, 2, 4),
        }
        buffer = RolloutBuffer(
            capacity=10,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=2,
            device="cpu",
            max_seq_len=max_seq_len,
            bptt_sequence_type=bptt_sequence_type,
            recurrent=True,
            hidden_state_architecture=hidden_state_arch,
        )

        # Create test data
        test_data = torch.arange(20).reshape(10, 2)  # 10 timesteps, 2 envs
        episode_done_indices = [[3, 7, 9], [4, 9]]  # Different episode endings per env

        sequences, max_length = buffer._get_complete_sequences(
            test_data, episode_done_indices
        )

        # Should have multiple sequences
        assert len(sequences) > 0
        assert max_length > 0

        # Check that sequences are torch tensors
        for seq in sequences:
            assert isinstance(seq, torch.Tensor)

    def test_convert_td_to_np_dict_method(self):
        """Test the _convert_td_to_np_dict method."""
        obs_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)

        buffer = RolloutBuffer(
            capacity=5,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=1,
            device="cpu",
        )

        # Create a test TensorDict
        test_td = TensorDict(
            {
                "observations": torch.randn(3, 2),
                "actions": torch.randint(0, 2, (3, 1)),
                "rewards": torch.randn(3),
                "hidden_states": TensorDict(
                    {
                        "h_actor": torch.randn(3, 1, 4),
                    },
                    batch_size=[3],
                ),
            },
            batch_size=[3],
        )

        # Convert to numpy dict
        np_dict = buffer._convert_td_to_np_dict(test_td)

        # Check that conversion worked
        assert isinstance(np_dict, dict)
        assert "observations" in np_dict
        assert "actions" in np_dict
        assert "rewards" in np_dict
        assert "hidden_states" in np_dict

        # Check that arrays are numpy arrays
        assert isinstance(np_dict["observations"], np.ndarray)
        assert isinstance(np_dict["actions"], np.ndarray)
        assert isinstance(np_dict["rewards"], np.ndarray)
        assert isinstance(np_dict["hidden_states"], dict)
        assert isinstance(np_dict["hidden_states"]["h_actor"], np.ndarray)

    def test_sequence_preparation_with_multiple_episodes_per_env(self):
        """Test sequence preparation with multiple episodes per environment."""
        obs_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        hidden_state_arch = {
            "h_actor": (1, 2, 4),
        }
        buffer = RolloutBuffer(
            capacity=20,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=2,
            device="cpu",
            max_seq_len=4,
            recurrent=True,
            hidden_state_architecture=hidden_state_arch,
        )

        # Add multiple episodes with different lengths
        for i in range(15):
            obs = np.array([[i, i + 1], [i + 2, i + 3]])
            action = np.array([i % 2, (i + 1) % 2])
            reward = np.array([1.0, -1.0])
            # Create episodes of different lengths for each environment
            done = np.array([i % 6 == 5, i % 4 == 3])
            value = np.array([0.5, -0.5])
            log_prob = np.array([0.1, 0.2])

            buffer.add(obs, action, reward, done, value, log_prob)

        # Compute returns
        last_value = np.array([0.8, 0.0])
        last_done = np.array([True, True])
        buffer.compute_returns_and_advantages(last_value, last_done)

        # Prepare sequence tensors
        buffer.prepare_sequence_tensors()

        # Should have created multiple sequences
        assert buffer.num_sequences > 2
        assert buffer.max_sequence_length <= 4

    def test_sequence_preparation_short_episodes(self):
        """Test sequence preparation with episodes shorter than max_seq_len."""
        obs_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        hidden_state_arch = {
            "h_actor": (1, 2, 4),
        }
        buffer = RolloutBuffer(
            capacity=10,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=1,
            device="cpu",
            max_seq_len=5,  # Longer than episodes
            recurrent=True,
            hidden_state_architecture=hidden_state_arch,
        )

        # Add short episodes
        for episode in range(3):
            for i in range(2):  # Each episode is only 2 steps
                obs = np.array([[episode * 10 + i, episode * 10 + i + 1]])
                action = np.array([i % 2])
                reward = np.array([1.0])
                done = np.array([i == 1])  # Episode ends after 2 steps
                value = np.array([0.5])
                log_prob = np.array([0.1])

                buffer.add(obs, action, reward, done, value, log_prob)

        # Compute returns
        last_value = np.array([0.0])
        last_done = np.array([True])
        buffer.compute_returns_and_advantages(last_value, last_done)

        # Prepare sequence tensors
        buffer.prepare_sequence_tensors()

        # Should have 3 sequences, each of length 2
        assert buffer.num_sequences == 3
        assert buffer.max_sequence_length == 2

    def test_minibatch_sequences_remainder_handling(self):
        """Test minibatch sequences with remainder sequences."""
        obs_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        hidden_state_arch = {
            "h_actor": (1, 2, 4),
        }
        buffer = RolloutBuffer(
            capacity=15,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=1,
            device="cpu",
            max_seq_len=2,
            recurrent=True,
            hidden_state_architecture=hidden_state_arch,
        )

        # Add 5 short episodes (will create 5 sequences)
        for episode in range(5):
            for i in range(2):
                obs = np.array([[episode * 10 + i, episode * 10 + i + 1]])
                action = np.array([i % 2])
                reward = np.array([1.0])
                done = np.array([i == 1])
                value = np.array([0.5])
                log_prob = np.array([0.1])

                buffer.add(obs, action, reward, done, value, log_prob)

        # Compute returns
        last_value = np.array([0.0])
        last_done = np.array([True])
        buffer.compute_returns_and_advantages(last_value, last_done)

        # Prepare sequence tensors
        buffer.prepare_sequence_tensors()

        # Get minibatches with batch_size=2 (should have 2 full batches + 1 remainder)
        batch_size = 2
        minibatch_generator = buffer.get_minibatch_sequences(batch_size)

        batch_sizes = []
        for padded_batch, unpadded_batch in minibatch_generator:
            # Count sequences in this batch by checking the batch dimension
            batch_sequences = padded_batch.batch_size[0] // buffer.max_sequence_length
            batch_sizes.append(batch_sequences)

        # Should have 2 batches of size 2 and 1 batch of size 1
        assert 2 in batch_sizes  # Full batches
        assert 1 in batch_sizes  # Remainder batch

    @pytest.mark.parametrize("recurrent", [True, False])
    def test_reset(self, recurrent):
        """Test buffer reset."""
        obs_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        hidden_state_arch = {
            "h_actor": (1, 2, 4),
        }
        buffer = RolloutBuffer(
            capacity=5,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=2,
            device="cpu",
            max_seq_len=3,
            recurrent=recurrent,
            hidden_state_architecture=hidden_state_arch,
        )

        # Add some data
        obs = np.array([[1.0, 2.0], [3.0, 4.0]])
        action = np.array([0, 1])
        reward = np.array([1.0, -1.0])
        done = np.array([False, True])
        value = np.array([0.5, -0.5])
        log_prob = np.array([0.1, 0.2])
        hidden_state = {"h_actor": torch.randn(1, 2, 4)}

        buffer.add(
            obs, action, reward, done, value, log_prob, hidden_state=hidden_state
        )

        # Prepare sequences
        buffer.compute_returns_and_advantages(
            np.array([0.0, 0.0]), np.array([True, True])
        )

        if recurrent:
            buffer.prepare_sequence_tensors()
        else:
            with pytest.raises(ValueError):
                buffer.prepare_sequence_tensors()

            return

        # Check that buffer has data
        assert buffer.pos > 0
        assert buffer.padded_data is not None
        assert buffer.unpadded_data is not None

        # Reset buffer
        buffer.reset()

        # Check that buffer is reset
        assert buffer.pos == 0
        assert buffer.full is False
        assert buffer.num_sequences is None
        assert buffer.max_sequence_length is None
        assert buffer.unpadded_slices is None
        assert buffer.padded_data is None
        assert buffer.unpadded_data is None

    def test_size_method(self):
        """Test size method."""
        obs_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        buffer = RolloutBuffer(
            capacity=5,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=3,
            device="cpu",
        )

        assert buffer.size() == 0

        # Add one timestep
        obs = np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
        action = np.array([0, 1, 0])
        reward = np.array([1.0, -1.0, 0.5])
        done = np.array([False, True, False])
        value = np.array([0.5, -0.5, 0.0])
        log_prob = np.array([0.1, 0.2, 0.3])

        buffer.add(obs, action, reward, done, value, log_prob)

        assert buffer.size() == 3  # 3 environments

    def test_serialization(self):
        """Test buffer serialization and deserialization."""
        obs_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        buffer = RolloutBuffer(
            capacity=5,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=2,
            device="cpu",
        )

        # Add some data
        obs = np.array([[1.0, 2.0], [3.0, 4.0]])
        action = np.array([0, 1])
        reward = np.array([1.0, -1.0])
        done = np.array([False, True])
        value = np.array([0.5, -0.5])
        log_prob = np.array([0.1, 0.2])

        buffer.add(obs, action, reward, done, value, log_prob)

        # Get state
        state = buffer.__getstate__()
        assert isinstance(state, dict)
        assert "buffer" in state
        assert "pos" in state
        assert "full" in state

        # Create new buffer and set state
        new_buffer = RolloutBuffer(
            capacity=5,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=2,
            device="cpu",
        )

        new_buffer.__setstate__(state)

        assert new_buffer.pos == buffer.pos
        assert new_buffer.full == buffer.full
        assert new_buffer.size() == buffer.size()

    def test_different_action_spaces(self):
        """Test with different action spaces."""
        obs_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)

        # Test with Box action space
        action_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        buffer = RolloutBuffer(
            capacity=5,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=2,
            device="cpu",
        )

        obs = np.array([[1.0, 2.0], [3.0, 4.0]])
        action = np.array([[0.5, -0.5], [1.0, -1.0]])
        reward = np.array([1.0, -1.0])
        done = np.array([False, True])
        value = np.array([0.5, -0.5])
        log_prob = np.array([0.1, 0.2])

        buffer.add(obs, action, reward, done, value, log_prob)

        assert buffer.pos == 1
        assert buffer.buffer["actions"].shape == (5, 2, 2)

        # Test with MultiDiscrete action space
        action_space = spaces.MultiDiscrete([2, 3])
        buffer = RolloutBuffer(
            capacity=5,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=2,
            device="cpu",
        )

        action = np.array([[0, 1], [1, 2]])
        buffer.add(obs, action, reward, done, value, log_prob)

        assert buffer.pos == 1
        assert buffer.buffer["actions"].shape == (5, 2, 2)

    def test_single_environment(self):
        """Test with single environment."""
        obs_space = spaces.Box(low=-1, high=1, shape=(3,), dtype=np.float32)
        action_space = spaces.Discrete(2)
        buffer = RolloutBuffer(
            capacity=5,
            observation_space=obs_space,
            action_space=action_space,
            num_envs=1,
            device="cpu",
        )

        obs = np.array([1.0, 2.0, 3.0])
        action = np.array(0)
        reward = np.array(1.0)
        done = np.array(False)
        value = np.array(0.5)
        log_prob = np.array(0.1)

        buffer.add(obs, action, reward, done, value, log_prob)

        assert buffer.pos == 1
        assert buffer.size() == 1
        assert buffer.buffer["observations"].shape == (5, 1, 3)
        assert buffer.buffer["actions"].shape == (5, 1, 1)
